#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "config.h"

#ifndef HAVE_GETOPT
#include "getopt.c"
#endif

#define BLOCKSIZE 1024

struct indext {
    off_t offset;
    long int length;
    time_t time;
};

char *getfield(char *line, int field);
static int indexcompare(const void *i, const void *j);

int main(int argc, char *argv[])
{
    char strip = 0;
    int c;
    long int size = 0, slide = 0;
    char *line = NULL;
    char *outfile = NULL;
    char *field;
    int opt;
    FILE *fileout;
    FILE *filein;
    struct tm timedate;
    struct indext *index = NULL;
    struct indext *indexptr = NULL;
    long int lines = 0;
    long int lineslide = 0;
    off_t startposn = 0;
    off_t offset = 0;
    off_t posn = 0;

    while ((opt = getopt(argc, argv, "o:shv")) != EOF) {
	switch (opt) {
	case 'o':
	    outfile = optarg;
	    break;

	case 's':
	    strip = 1;
	    break;

	case 'h':
	    printf("elfsort [-s] [-o output-file] [file]\n");
	    exit(0);

	case 'v':
	    printf("elfsort - " PACKAGE " " VERSION "\n");
	    exit(0);
	}
    }

    /* open up file for output */
    if (outfile) {
	if (!(fileout = fopen(outfile, "w"))) {
	    perror(outfile);
	    exit(1);
	}
    }
    else
	fileout = stdout;


    /* open up and read input file */
    opt = optind;
    if (opt == argc)
	filein = stdin;
    else
	filein = fopen(argv[opt], "r");
    if (!filein) {
	perror(argv[opt]);
	exit(1);
    }


    /* we loop through the file, and create an index of
     * each line's start and length, as well as the unix
     * date of that line. afterwards we'll sort this index
     * by date, then output the lines pulling out in the
     * correct order. */

    /* get original file position */
    startposn = ftell(filein);

    /* get each line of text until EOF */
    while ((c = fgetc(filein)) != EOF) {
	/* enough space in buffer? no, make some more */
	if (slide >= size) {
	    size += BLOCKSIZE;
	    if (!(line = (char *) realloc(line, size))) {
		fprintf(stderr, "Memory allocation error\n");
		exit(1);
	    }
	}

	/* save character to buffer */
	line[slide++] = (char) c;

	/* reached an end of line? if so, deal with the line */
	if (c == 10) {

	    /* the line may be terminated with a CR, if so chop it off
	     * otherwise replace the LF with a NULL to indicate end of string */
	    if ((slide > 0) && (line[slide - 1] == 13))
		line[slide - 2] = 0;
	    else
		line[slide - 1] = 0;

	    /* get the unix date from the line */
	    field = getfield(line,4);
	    timedate.tm_sec = (field[19]-'0')*10+(field[20]-'0');
	    timedate.tm_min = (field[16]-'0')*10+(field[17]-'0');
	    timedate.tm_hour = (field[13]-'0')*10+(field[14]-'0');
	    timedate.tm_mday = (field[1]-'0')*10+(field[2]-'0');
	    switch (field[4]) {
	    case 'J' : {if(field[5] == 'a') timedate.tm_mon = 0; else if(field[6] == 'l') timedate.tm_mon = 6; else timedate.tm_mon = 5; break;}
	    case 'F' : {timedate.tm_mon = 1; break;}
	    case 'M' : {if(field[6] == 'y') timedate.tm_mon = 4; else timedate.tm_mon = 2; break;}
	    case 'A' : {if(field[5] == 'p') timedate.tm_mon = 3; else timedate.tm_mon = 7; break;}
	    case 'S' : {timedate.tm_mon = 8; break;}
	    case 'O' : {timedate.tm_mon = 9; break;}
	    case 'N' : {timedate.tm_mon = 10; break;}
	    case 'D' : {timedate.tm_mon = 11; break;}
	    }
	    timedate.tm_year = (field[8]-'0')*1000+(field[9]-'0')*100+(field[10]-'0')*10+(field[11]-'0')-1900;
	    timedate.tm_isdst = -1;
	    free(field);
	    
	    field = getfield(line,5);
	    timedate.tm_hour -= atoi(field);
	    free(field);

	    /* add an entry to the index */
	    if (!(index = (struct indext *) realloc(index, (lines+1)*sizeof(struct indext)))) {
		fprintf(stderr, "Memory allocation error\n");
		exit(1);
	    }
	    posn = ftell(filein);
	    indexptr = index + lines;
	    indexptr->offset = offset;
	    indexptr->length = posn - offset;
	    indexptr->time = mktime(&timedate);
	    offset = posn;
	    lines++;

	    /* should we strip corrupt lines? */
	    if (strip && !indexptr->time)
		indexptr->length = 0;

	    /* clean up */
	    free(line);
	    line = NULL;
	    slide = 0;
	    size = 0;
	}
    }
    if (line)
	free(line);

    /* sort the actual array */
    qsort(index, lines, sizeof(struct indext), indexcompare);

    /* output the logs in their new order */
    indexptr = index;
    lineslide = 0;
    while (lineslide < lines) {
#if HAVE_FSEEKO
	fseeko(filein, indexptr->offset, SEEK_SET);
#elif HAVE_FSEEK
	fseek(filein, (long int)indexptr->offset, SEEK_SET);
#else
#error "Neither fseeko nor fseek found in libc. Aborting."
#endif
	if (!(line = (char *) malloc(indexptr->length+1))) {
	    fprintf(stderr, "Memory allocation error\n");
	    exit(1);
	}
	if(indexptr->length != fread(line,1,indexptr->length,filein)) {
	    perror("read error");
	    exit(1);
	}
	if(indexptr->length != fwrite(line,1,indexptr->length,fileout)) {
	    perror("write error");
	    exit(1);
	}
	free(line);
	line = NULL;
	indexptr++;
	lineslide++;
    }


    /* close input file (but only if the file was a real file) */
    if (filein != stdin)
	fclose(filein);
    filein = NULL;

    /* close out output file */
    if (fileout != stdout)
	fclose(fileout);

    exit(0);
}

/* comparison function */
static int indexcompare(const void *i, const void *j)
{
    if (((struct indext *)i)->time > ((struct indext *)j)->time)
	return(1);
    if (((struct indext *)i)->time < ((struct indext *)j)->time)
	return(-1);
    return(0);
}


/* get field number in a line */
/* buffer returned is size + 127 bytes so that it may be overwritten by the calling routine */
char *getfield(char *line, int field)
{
    char *buf;
    int start = 0, end = 0, slide = 0;
    int ws = 0;
    int br = 0;

    /* if ws is 0, we are finding text. if ws is not 0, we are finding whitespace */
    /* if br is not 0, we are inside a set of brackets */
    while (field && line[slide] && (line[slide] != 13)) {

	/* deal with brackets */
	if (line[slide] == '"') {
	    if (br)
		br = 0;
	    else
		br = 1;
	}

	/* find whitespace */
	switch (line[slide]) {
	case 8:
	case ' ':{
		if (!br && ws) {
		    ws = 0;
		}
		break;
	    }
	default:{
		if (!ws) {
		    field--;
		    ws = 1;
		    start = slide;
		}
	    }
	}
	slide++;
    }

    /* we found the start, now find the end */
    while (!end && line[slide] && (line[slide] != 13)) {

	/* find whitespace */
	switch (line[slide]) {
	case '"':{
		if (br)
		    br = 0;
		else
		    br = 1;
		break;
	    }
	case 8:
	case ' ':{
		if (!br) {
		    end = slide;
		}
		break;
	    }
	}
	slide++;
    }

    /* if !end, we found the last field */
    if (!end)
	end = slide;

    /* if field not found, return "" */
    if (field)
	start = end = 0;

    /* create a buffer for the field and return it - buffer contains an extra 127 bytes */
    if (!(buf = (char *) malloc(end - start + 128)))
	return (NULL);
    strncpy(buf, line + start, end - start);
    buf[end - start] = 0;
    return (buf);
}

